/*
 * RemCon.c
 *
 * Универсальный пульт дистанционного управления на базе nRF24L01+.
 *
 * Author: Погребняк Дмитрий
 * 
 * Помещённый здесь код является свободным. Т.е. допускается его свободное некоммерческое использование при условии указания ссылки на автора (Погребняк Дмитрий, http://aterlux.ru/).
 * Code included here is free to use. It can be used for free for non-commercial purposes, as long as link to author (Pogrebnyak Dmitry, http://aterlux.ru/) is included.
 * 
 * Фьюзы/Fuses (ATmega48PA): extended 0xFF, high 0xD6, low 0x62
 * Отличие от настроек по-умолчанию: включен BOD на 1.8 Вольт, включена опция EESAVE
 */ 

#define F_CPU 1000000UL
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include "nRF24L01.h"
#include "eeprom.h"

#define EEPROM_REMOTE_ADDRESS 0
#define EEPROM_CHANNEL 5
#define EEPROM_RETR 6
#define EEPROM_FLAGS 7
#define EEPROM_REPEAT_PAUSE 8
#define EEPROM_DBLCLK_DELAY 9
#define EEPROM_LONG_HOLD 10
#define EEPROM_NO_REPLY_LIMIT 11

#define FLAG_SPEED_2MBSP 0
#define FLAG_CRC_2BYTES 1
#define FLAG_POWER 2
#define FLAG_ADRLEN 4

#define DEFAULT_CHANNEL 2
#define DEFAULT_FLAGS ((1 << FLAG_SPEED_2MBSP) | (0 << FLAG_CRC_2BYTES) | (0b11 << FLAG_POWER) | (0b11 < FLAG_ADRLEN))
#define DEFAULT_RETR 0b00000010
#define DEFAULT_REPEAT_PAUSE 12
#define DEFAULT_DBLCLK_DELAY 45
#define DEFAULT_LONG_HOLD 120
#define DEFAULT_NO_REPLY_LIMIT 40


const PROGMEM uint8_t default_remote_address[5]  = {0xC2, 0xC2, 0xC2, 0xC2, 0xC2};


#define BUTTON_PIN PINB
#define BUTTON_PORT PORTB
#define BUTTON_DDR DDRB
#define BUTTON_0 (1 << 4) // Верхняя
#define BUTTON_1 (1 << 3) // Нижняя
#define BUTTON_2 (1 << 7) // Левая
#define BUTTON_3 (1 << 5) // Правая
#define BUTTON_MASK (BUTTON_0 | BUTTON_1 | BUTTON_2 | BUTTON_3)


volatile uint8_t dblclk[4];
volatile uint8_t repeat_countdown, do_repeat;
volatile uint8_t hold_count;

uint8_t repeat_pause;
uint8_t dblclk_delay;
uint8_t long_hold_threshold;
uint8_t no_reply_max;
uint8_t was_power_down;

uint8_t long_hold, dblclk_mask, changed_mask, change_cnt;
uint8_t config_reg;

uint8_t radio_start_module() {
  radio_deassert_ce();
  uint8_t flags = eeprom_read(EEPROM_FLAGS, DEFAULT_FLAGS);
  if ((((flags >> FLAG_ADRLEN) & 3) == 0) || (flags & 0xC0)) 
    flags = DEFAULT_FLAGS;
    
  config_reg = (1 << EN_CRC) | ((flags & (1 << FLAG_CRC_2BYTES)) ? (1 << CRCO) : 0);
  for(uint8_t cnt = 100;;) {
    radio_writereg(CONFIG, config_reg); // Питание выключено
    if (radio_readreg(CONFIG) == config_reg) 
      break;
    if (!cnt--)
      return 0;
    _delay_ms(1);
  }
  radio_writereg(EN_AA, 0); // выключение автоподтверждений
  radio_writereg(EN_RXADDR, (1 << ERX_P0)); // включение канала 0 
  radio_writereg(SETUP_AW, (flags >> FLAG_ADRLEN) & 3); // выбор длины адреса 5 байт
  radio_writereg(SETUP_RETR, eeprom_read(EEPROM_RETR, DEFAULT_RETR)); // до 2 повторов передачи, 250мкс между ними
  radio_writereg(RF_CH, eeprom_read(EEPROM_CHANNEL, DEFAULT_CHANNEL)); // Выбор частотного канала
  radio_writereg(RF_SETUP, (((flags >> FLAG_POWER) & 3) << RF_PWR) | ((flags & (1 << FLAG_SPEED_2MBSP)) ? (1 << RF_DR_HIGH) : 0));
  
  uint8_t addr[5];
  if (!eeprom_read_buf(EEPROM_REMOTE_ADDRESS, 5, &addr[0])) 
    memcpy_P(&addr, &default_remote_address, 5);
    
  radio_writereg_buf(RX_ADDR_P0, &addr[0], 5);
  radio_writereg_buf(TX_ADDR, &addr[0], 5);

  radio_writereg(RX_PW_P0, 0);
  radio_writereg(RX_PW_P1, 32);
  radio_writereg(DYNPD, (1 << DPL_P0)); // включение произвольной длины данных для каналов 0 и 1
  radio_writereg(FEATURE, (1 << EN_DPL)); // разрешение произвольной длины пакета данных

  uint8_t st = radio_cmd(NOP);
  radio_writereg(STATUS, st);
  uint8_t protect = 5;
  while (!(radio_readreg(FIFO_STATUS) & (1 << TX_EMPTY)) && --protect) 
    radio_cmd(FLUSH_TX);
  if (!protect)
    return 0;
  if ((st & (7 << RX_P_NO)) != (7 << RX_P_NO))
    radio_cmd(FLUSH_RX);
  return 1;
}

void power_down() {
  was_power_down = 1;
  radio_writereg(CONFIG, config_reg); // отключение питания
  cli(); // отключение прерываний
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  sleep_enable();
  uint8_t x = (MCUCR & ~(1 << BODSE)) | (1 << BODS); // подготовка бит
  MCUCR = x | (1 << BODSE); // процедура отключения BOD
  MCUCR = x;
  sei(); // включение прерываний
  sleep_cpu(); // вход в режим сна  
  BUTTON_PORT |= BUTTON_MASK; // включаем pull-up на всех кнопках
  set_sleep_mode(SLEEP_MODE_PWR_SAVE);
}

EMPTY_INTERRUPT(PCINT1_vect);

EMPTY_INTERRUPT(PCINT0_vect);

ISR(TIMER2_OVF_vect) {
  uint8_t t;
  t = dblclk[0]; if (t) dblclk[0] = t - 1;
  t = dblclk[1]; if (t) dblclk[1] = t - 1;
  t = dblclk[2]; if (t) dblclk[2] = t - 1;
  t = dblclk[3]; if (t) dblclk[3] = t - 1;
  t = repeat_countdown; 
  if (t) {
    t--;
    repeat_countdown = t;
    if (!t) {
      do_repeat = 1;
    }
  }  
  t = hold_count;
  if (t < 255) {
    hold_count = t + 1;
  }
}


int main(void)
{
  PORTB = 0b11111111; // Подтяжка на всех портах
  PORTD = 0b11111111;
  PORTC = 0b00111111;
  ACSR |= (1 << ACD); // Отключение компаратора
  radio_init();
  while (!radio_start_module()) {
    power_down();
  }
  
  repeat_pause = eeprom_read(EEPROM_REPEAT_PAUSE, DEFAULT_REPEAT_PAUSE);
  dblclk_delay = eeprom_read(EEPROM_DBLCLK_DELAY, DEFAULT_DBLCLK_DELAY);
  long_hold_threshold = eeprom_read(EEPROM_LONG_HOLD, DEFAULT_LONG_HOLD);
  no_reply_max = eeprom_read(EEPROM_NO_REPLY_LIMIT, DEFAULT_NO_REPLY_LIMIT);
  PCMSK0 = BUTTON_MASK;
  PCMSK1 = IRQ;
  PCICR = (1 << PCIE1) | (1 << PCIE0);
  TCCR2A = 0;
  TCCR2B = 0b00000100; // прескалер 1/128, переполнение 1МГц / 64 / 256 = 61 раз в секунду (период 16.384 мс)
  TIMSK2 = (1 << TOIE2);
  
  set_sleep_mode(SLEEP_MODE_PWR_SAVE);
  sleep_enable();
  was_power_down = 1;
  uint8_t cur_buttons_state = 0;
  uint8_t in_transmit = 0;
  uint8_t no_reply_count = 0;
  for(;;) {
    if (radio_is_interrupt()) {
      sei();
      uint8_t status = radio_cmd(NOP);
      radio_writereg(STATUS, status); // Просто запишем регистр обратно, тем самым сбросив биты прерываний
      if (status & ((1 << TX_DS) | (1 << MAX_RT))) { // Завершена передача успехом, или нет,
        if (status & (1 << MAX_RT)) { // Если достигнуто максимальное число попыток
          radio_cmd(FLUSH_TX); // Удалим последний пакет из очереди
          if (no_reply_count < 255)
            no_reply_count++;
        } else {
          no_reply_count = 0;    
        }        
        in_transmit = 0;
      }
      uint8_t protect = 4; // В очереди FIFO не должно быть более 3 пакетов. Если больше, значит что-то не так
      while (((status & (7 << RX_P_NO)) != (7 << RX_P_NO)) && protect--) { // Пока в очереди есть принятый пакет
        radio_cmd(FLUSH_RX); // во всех случаях выкидываем пришедший пакет.
        status = radio_cmd(NOP);
      }
      cli();
    } else if (!in_transmit) { 
      sei();
      // Обработка кнопок ведётся только пока не происходит передача, чтобы исключить вариант, когда два пакета передаются одновременно.
      uint8_t btns = 0; // Текущее состояние кнопок
      if (!(BUTTON_PIN & BUTTON_0)) btns |= (1 << 0);
      if (!(BUTTON_PIN & BUTTON_1)) btns |= (1 << 1);
      if (!(BUTTON_PIN & BUTTON_2)) btns |= (1 << 2);
      if (!(BUTTON_PIN & BUTTON_3)) btns |= (1 << 3);
      uint8_t do_send = 0;
      if (btns != cur_buttons_state) {
        changed_mask = btns ^ cur_buttons_state;
        cur_buttons_state = btns;
        change_cnt++;
        hold_count = 0;
        long_hold = 0;
        do_send = 1;
        no_reply_count = 0;
        uint8_t bmsk = 1;
        for (uint8_t i = 0; i < 4; i++) {
          if (changed_mask & bmsk) { // Если состояние кнопки изменилось
            if (btns & bmsk) { // Если кнопка была нажата
              if (dblclk[i]) { // Ещё действует счётчик двойного нажатия для кнопки
                dblclk_mask |= bmsk;  // Устанавливается флаг двойного щелчка
              } else {
                dblclk_mask &= ~bmsk; // Сбрасывается флаг двойного щелчка
              }
              dblclk[i] = dblclk_delay; // Устанавливается обратный счётчик двойного щелчка на эту кнопку
            }
          }
          bmsk <<= 1;
        }        
      } else if (do_repeat) {
        do_repeat = 0;
        if (no_reply_max && (no_reply_count > no_reply_max)) {
          // Если кнопки остаются нажатыми, а подтверждений о доставке стабильно нет в течение заданного интервала,
          // То прикинемся что кнопки не нажаты, чтобы перейти в спящий режим ниже
          btns = 0;
          dblclk[0] = 0;
          dblclk[1] = 0;
          dblclk[2] = 0;
          dblclk[3] = 0;
          uint8_t pressed = BUTTON_PIN | ~BUTTON_MASK; // нули на тех битах местах, где кнопки зажаты, остальные единицы.
          if (pressed == (uint8_t)~BUTTON_MASK)  // Исключительная ситуация: если зажаты все кнопки
            pressed |= BUTTON_0; // оставляем pull-up на нулевой (верхней) кнопке, чтобы можно было выйти из спящего режима
          BUTTON_PORT &= pressed; // выключаем pull-up на зажатых кнопках, для экономии батареи
        } else {
          do_send = 1;
        }        
      }
      if (do_send) {
        if (was_power_down) { // Если был в режиме отключения питания до этого, то нужно включить и выждать минимум 1.5 мс до запуска осциллятора
          radio_writereg(CONFIG, config_reg | (1 << PWR_UP)); // primTX
          TCNT0 = 230; // 1.6мс до переполнения
          cli();
          do_repeat = 0;
          repeat_countdown = 1; // установка переменной do_repeat при следующем переполнении
          while (!do_repeat) { // Ждём пока произойдёт переполнение в спящем режиме
            sei();
            sleep_cpu();
            cli();
          }
          sei();
          do_repeat = 0;
          was_power_down = 0;
        }
        // Отправка информации
        if (btns) { // Если нажаты кнопки
          if (hold_count > long_hold_threshold) // Проверим и установим флаг долгого удержания
            long_hold = 1;
          repeat_countdown = repeat_pause; // Обратный отсчёт до повтора
        } else {
          repeat_countdown = 0; // Если не нажаты, то одиночная отправка, без повторов
          do_repeat = 0;
        }
        uint8_t buf[5];
        buf[0] = btns | (dblclk_mask << 4);
        buf[1] = changed_mask | ((change_cnt & 7) << 4) | (long_hold ? 0x80 : 00);
        in_transmit = 1;
        
        radio_write_buf(W_TX_PAYLOAD, &buf[0], 2);
        radio_assert_ce();
        _delay_us(15);
        radio_deassert_ce();
      } else if (!btns && !dblclk[0] && !dblclk[1] && !dblclk[2] && !dblclk[3]) {
        power_down(); // Если кнопки не нажаты и отправлять ничего не надо, то переходим в режим сна
      }
      cli();
    } else {
      sei();
      sleep_cpu();
      cli();
    }
  }
}