﻿/*
 * eeprom.c
 *
 * Модуль для работы с EEPROM
 *
 * Author: Погребняк Дмитрий
 * 
 * Помещённый здесь код является свободным. Т.е. допускается его свободное использование для любых целей, включая коммерческие, при условии указания ссылки на автора (Погребняк Дмитрий, http://aterlux.ru/).
 * Code included here is free to use. It can be used for free for any purpose, including commercial use, as long as link to author (Pogrebnyak Dmitry, http://aterlux.ru/) is included.
 */ 

#include <avr/io.h>
#include <avr/interrupt.h>
#include "eeprom.h"


uint8_t eeprom_read(eeprom_addr_t address, uint8_t return_if_0xFF) {
  while(EECR & (1<<EEPE))   // Ожидание окончания записи EEPROM
  ; 
  uint8_t oldSREG = SREG;   // Сохранение флага прерываний
  cli();                    // Отключение прерываний
  while(EECR & (1<<EEPE))   // Повтороное ожидание окончания записи EEPROM, на случай если запись была начата в прерывании
  ;
  #ifdef EEAR
    EEAR = address;
  #else
    EEARL = address;
  #endif
  EECR |= (1<<EERE);        // Выбор операции - чтение
  uint8_t d = EEDR;         // Чтение EEPROM
  SREG = oldSREG;           // Восстановление флага прерываний
  return (d == 0xFF) ? return_if_0xFF : d;
}
  
void eeprom_write(eeprom_addr_t address, uint8_t data) {
  while(EECR & (1<<EEPE))   // Ожидание окончания предыдущей операции записи EEPROM
  ;
  uint8_t oldSREG = SREG;   // Сохранение флага прерываний
  cli();                    // Отключение прерываний
  while(EECR & (1<<EEPE))   // Повтороное ожидание окончания записи EEPROM, на случай если запись была начата в прерывании
  ;
  #ifdef EEAR
    EEAR = address;
  #else
    EEARL = address;
  #endif
  EECR |= (1<<EERE);        // Выбор операции - чтение
  uint8_t d = EEDR;         // Чтение EEPROM
  if (d != data) {          // Запись проивзодится только если записываемые данные отличаются от уже записанных
    EEDR = data;            // Загрузка данных для записи
    EECR |= (1<<EEMPE);     // Разрешение записи
    EECR |= (1<<EEPE);      // Выбор операции - запись
  }
  SREG = oldSREG;           // Восстановление флага прерываний
}

int16_t eeprom_read_int16(eeprom_addr_t address, int16_t return_if_0xFFFF) {
  int16_t w = eeprom_read(address, 0xFF) | (eeprom_read(address + 1, 0xFF) << 8);
  return (w == -1) ? return_if_0xFFFF : w;
}

void eeprom_write_int16(eeprom_addr_t address, int16_t data) {
  eeprom_write(address, data);
  eeprom_write(address + 1, data >> 8);
}

uint8_t eeprom_read_buf(eeprom_addr_t address, uint8_t count, uint8_t * buf) {
  uint8_t r = 0xFF;
  while (count--) {
    uint8_t b = eeprom_read(address++, 0xFF);
    r &= b;
    *(buf++) = b;
  }
  return r != 0xFF;
}  

void eeprom_write_buf(eeprom_addr_t address, uint8_t * buf, uint8_t count) {
  while (count--) 
    eeprom_write(address++, *(buf++));
} 
