﻿/*
 * nRF24L01.c
 *
 * Модуль для работы с чипами радиосвязи nRF24L01+
 *
 * Author: Погребняк Дмитрий
 * 
 * Помещённый здесь код является свободным. Т.е. допускается его свободное использование для любых целей, включая коммерческие, при условии указания ссылки на автора (Погребняк Дмитрий, http://aterlux.ru/).
 * Code included here is free to use. It can be used for free for any purpose, including commercial use, as long as link to author (Pogrebnyak Dmitry, http://aterlux.ru/) is included.
 */ 

#include "nRF24L01.h"
#include "spi.h"


// Поскольку функции для работы с csn не предполагается использовать в иных файлах, их можно объявить static

// Выбирает активное состояние (низкий уровень) на линии CSN
inline static void csn_assert() {
  CSN_PORT  &= ~CSN; // Установка низкого уровня на линии CSN
}

// Выбирает неактивное состояние (высокий уровень) на линии CSN
inline static void csn_deassert() {
  CSN_PORT |= CSN; // Установка высокого уровня на линии CSN
}

void radio_init() {
  CSN_PORT |= CSN;
  CSN_DDR |= CSN;
  
  spi_init();
  
  CE_PORT &= ~CE;
  CE_DDR |= CE;
  IRQ_DDR &= ~IRQ;
  IRQ_PORT |= IRQ;
  UBRR0 = 0;
  
  UCSR0C = (1 << UMSEL01) | (1 << UMSEL00);
  UCSR0B = (1 << RXEN0) | (1 << TXEN0);
  
  UBRR0 = 0;
}

uint8_t radio_readreg(uint8_t reg) {
  csn_assert();
  spi_send_recv(reg | R_REGISTER);
  uint8_t answ = spi_send_recv(0xFF);
  csn_deassert();
  return answ;
}

uint8_t radio_writereg(uint8_t reg, uint8_t val) {
  csn_assert();
  uint8_t status = spi_send_recv(reg | W_REGISTER);
  spi_send_recv(val);
  csn_deassert();
  return status;
}

uint8_t  radio_write_buf(uint8_t cmd, uint8_t * buf, uint8_t count) {
  csn_assert();
  uint8_t status = spi_send_recv(cmd);
  while (count--) {
    spi_send_recv(*(buf++));
  }
  csn_deassert();
  return status;
}

uint8_t  radio_writereg_buf(uint8_t reg, uint8_t * buf, uint8_t count) {
  return radio_write_buf(reg | W_REGISTER, buf, count);
}

uint8_t radio_read_buf(uint8_t cmd, uint8_t * buf, uint8_t count) {
  csn_assert();
  uint8_t status = spi_send_recv(cmd);
  while (count--) {
    *(buf++) = spi_send_recv(0xFF);
  }
  csn_deassert();
  return status;
}

uint8_t  radio_readreg_buf(uint8_t reg, uint8_t * buf, uint8_t count) {
  return radio_read_buf(reg | R_REGISTER, buf, count);
}

uint8_t radio_read_rx_payload_width() {
  csn_assert();
  spi_send_recv(R_RX_PL_WID);
  uint8_t answ = spi_send_recv(0xFF);
  csn_deassert();
  return answ;
}

uint8_t radio_cmd(uint8_t cmd) {
  csn_assert();
  uint8_t status = spi_send_recv(cmd);
  csn_deassert();
  return status;
}

void radio_assert_ce() {
  CE_PORT |= CE;
}

void radio_deassert_ce() {
  CE_PORT &= ~CE;
}

uint8_t radio_is_interrupt() {
  return (IRQ_PIN & IRQ) ? 0 : 1;
}
