﻿/*
 * spi.c
 *
 * Модуль для работы с шиной SPI, посредством аппаратного интерфейса SPI, USART в режиме SPI-мастер, или программной реализацией SPI
 *
 * Author: Погребняк Дмитрий
 * 
 * Помещённый здесь код является свободным. Т.е. допускается его свободное использование для любых целей, включая коммерческие, при условии указания ссылки на автора (Погребняк Дмитрий, http://aterlux.ru/).
 * Code included here is free to use. It can be used for free for any purpose, including commercial use, as long as link to author (Pogrebnyak Dmitry, http://aterlux.ru/) is included.
 */ 

#include "spi.h"

#ifdef SPI_HARDWARE
// Аппаратный SPI

void spi_init() {
  SPI_DDR |= SPI_MOSI | SPI_SCK | SPI_SS;
  SPI_SPCR = (1 << SPE) | (SPI_DORD << DORD) | (1 << MSTR) | (SPI_CPOL << CPOL) | (SPI_CPHA << CPHA) | ((((SPI_SPEED_SCALE - 1) >> 1) & 3) << SPR0);
  SPI_SPSR = (((SPI_SPEED_SCALE - 1) & 1) << SPI2X);
}

uint8_t spi_send_recv(uint8_t data) {
  SPI_SPDR = data;
  while (!(SPI_SPSR & (1 << SPIF)));
  return SPI_SPDR;
}

#elif defined SPI_USART
// USART в режиме SPI-master

// Инициализация интерфейса USART в режиме SPI-master
void spi_init() {
  USART_UBRR = 0; 
  USART_DDR |= USART_XCK;
  // Исходим из предположения что расположение битов во всех USART одинаково...
  USART_UCSRC = (1 << UMSEL01) | (1 << UMSEL00) | (SPI_DORD << UDORD) | (SPI_CPHA << UCPHA) | (SPI_CPOL << UCPOL0); // выбор режима SPI-master
  UCSR0B = (1 << RXEN0) | (1 << TXEN0); // Включение приёмника и передатчика
  UBRR0 = 1;  // Выбор частоты интерфейса 1/4 от частоты ЦП
}

// Передаёт и принимает 1 байт по SPI, возвращает полученное значение
uint8_t spi_send_recv(uint8_t data) {
  USART_UDR = data;
  while (!(USART_UCSRA & (1 << RXC0)));
  return USART_UDR; 
}

#elif defined SPI_SOFTWARE
// Программный SPI

// Инициализация программного интерфейса SPI
void spi_init() {
  SWSPI_MOSI_DDR |= SWSPI_MOSI;
  
  SWSPI_MISO_DDR &= ~SWSPI_MISO;
  SWSPI_MISO_PORT |= SWSPI_MISO; // Включение pull-up на линии MISO
  
  SWSPI_SCK_DDR |= SWSPI_SCK_DDR;
  #if SPI_CPOL == 1 
    SWSPI_SCK_PORT |= SWSPI_SCK_DDR;
  #else
    SWSPI_SCK_PORT &= ~SWSPI_SCK_DDR;
  #endif
}

// Задержка обмена для низких скоростей
#if SPI_SPEED_SCALE > 3
  #define swspi_halfpause() for (uint8_t i = ((1 << (SPI_SPEED_SCALE - 2)) - 3); i; i--) {}
#else
  #define swspi_halfpause() 
#endif

static inline void swspi_setup(uint8_t data) {
 #if SPI_DORD == 0 
  if (data & 0x80) 
 #else
  if (data & 0x01) 
 #endif
    SWSPI_MOSI_PORT |= SWSPI_MOSI; // передача единички
  else
    SWSPI_MOSI_PORT &= ~SWSPI_MOSI; // передача нулика
}

static inline uint8_t swspi_shift_and_sample(uint8_t data) {
 #if SPI_DORD == 0 
  data <<= 1;
  if (SWSPI_MISO_PIN & SWSPI_MISO)
    data |= 1;
 #else
  data >>= 1;
  if (SWSPI_MISO_PIN & SWSPI_MISO)
    data |= 0x80;
 #endif 
  return data;
}

#if SPI_CPOL == 0
  #define swspi_sck_leading() SWSPI_SCK_PORT |= SWSPI_SCK
  #define swspi_sck_trailing() SWSPI_SCK_PORT &= ~SWSPI_SCK
#else
  #define swspi_sck_leading() SWSPI_SCK_PORT &= ~SWSPI_SCK
  #define swspi_sck_trailing() SWSPI_SCK_PORT |= SWSPI_SCK
#endif

// Передаёт и принимает 1 байт по SPI, возвращает полученное значение
uint8_t spi_send_recv(uint8_t data) {
  // От старшего к младшему
  for (uint8_t i = 8; i; i--) {
   #if SPI_CPHA != 0
    swspi_sck_leading();
   #endif
    swspi_setup(data);
    swspi_halfpause();
   #if SPI_CPHA == 0
    swspi_sck_leading();
   #else
    swspi_sck_trailing();
   #endif
    data = swspi_shift_and_sample(data);
    swspi_halfpause();
   #if SPI_CPHA == 0
    swspi_sck_trailing();
   #endif
  }
  return data;
}


#endif