﻿/*
 * spi.h
 *
 * Модуль для работы с шиной SPI, посредством аппаратного интерфейса SPI, USART в режиме SPI-мастер, или программной реализацией SPI
 *
 * Author: Погребняк Дмитрий
 * 
 * Помещённый здесь код является свободным. Т.е. допускается его свободное использование для любых целей, включая коммерческие, при условии указания ссылки на автора (Погребняк Дмитрий, http://aterlux.ru/).
 * Code included here is free to use. It can be used for free for any purpose, including commercial use, as long as link to author (Pogrebnyak Dmitry, http://aterlux.ru/) is included.
 */ 


#ifndef SPI_H_
#define SPI_H_

#include <avr/io.h>


/* Выбор типа SPI */

// Определить только одно из SPI_HARDWARE, SPI_USART, SPI_SOFTWARE
//#define SPI_HARDWARE // Аппаратный SPI интерфейс
#define SPI_USART // Аппаратный USART в качестве SPI-мастера, только для микроконтроллеров, где SPI может работать в режиме SPI-master!
//#define SPI_SOFTWARE // Программный SPI интерфейс

/* Общие настройки SPI */

#define SPI_CPOL 0 // Полярность: 0 - начало импульса - нарастающий фронт, окончание - спадающий; 1 - начало импульса - спад, окончание - нарастающий фронт
#define SPI_CPHA 0 // Фаза: 0 - по началу импульса - чтение значений, по окончанию - выставление (первый передаваемый бит выставлен заранее); 1 - по началу импульса - выставление значений, по окончанию - чтение

#define SPI_DORD 0 // Последовательность бит в байте: 0 - сначала старший бит; 1 - сначала младший.
#define SPI_SPEED_SCALE 1 // от 1 до 8: скорость SPI в 2 в степени SPI_SPEED_SCALE раз меньше, чем скорость CPU, т.е. 1 - 1/2, 2  - 1/4, 3 - 1/8 ... 8 - 1/256

/* Частные, в зависимости от типа, настройки SPI */

#ifdef SPI_HARDWARE

  #define SPI_DDR DDRB  // Порт на котором расположены выводы SPI

  #define SPI_SS (1 << 2)      // Маска регистра порта SPI_DDR для вывода SS
  #define SPI_MOSI (1 << 3)    // Маска регистра порта SPI_DDR для вывода MOSI
  #define SPI_SCK (1 << 5)     // Маска регистра порта SPI_DDR для вывода SCK
  
  #define SPI_SPCR SPCR // Регистр управления SPCR используемого SPI-интерфейса
  #define SPI_SPSR SPSR // Регистр состояния SPSR используемого SPI-интерфейса 
  #define SPI_SPDR SPDR // Регистр данных SPDR используемого SPI-интерфейса

  #if (defined SPI_USART) || (defined SPI_SOFTWARE)
    #error "different SPI types are defined simultaneously. Define only one of the SPI_HARDWARE, SPI_USART, SPI_SOFTWARE in spi.h"
  #endif
#elif defined SPI_USART

  #define USART_DDR DDRD // Порт, на котором находятся выводы USART

  #define USART_XCK (1 << 4)    // Маска регистра порта USART для вывода XCK
  
  #define USART_UBRR UBRR0   // Регистр настройки скорости используемого USART
  #define USART_UDR UDR0     // Регистр данных используемого USART
  #define USART_UCSRA UCSR0A // Регистр состояния и управления A используемого USART
  #define USART_UCSRB UCSR0B // Регистр состояния и управления B используемого USART
  #define USART_UCSRC UCSR0C // Регистр состояния и управления C используемого USART
  
  #define UCPHA 1 // номер бита UCPHAx в регистре UCSRxC 
  #define UDORD 2 // номер бита UDORDx в регистре UCSRxC 
  
  #if (defined SPI_SOFTWARE)
    #error "different SPI types are defined simultaneously. Define only one of the SPI_HARDWARE, SPI_USART, SPI_SOFTWARE in spi.h"
  #endif
#elif defined SPI_SOFTWARE
  // Регистры порта и выводы используемые для сигналов MOSI, MISO, SCK

  #define SWSPI_MOSI_DDR DDRB
  #define SWSPI_MOSI_PORT PORTB
  #define SWSPI_MOSI (1 << 6)

  #define SWSPI_MISO_DDR DDRB
  #define SWSPI_MISO_PORT PORTB
  #define SWSPI_MISO_PIN PINB
  #define SWSPI_MISO (1 << 7)

  #define SWSPI_SCK_DDR DDRB
  #define SWSPI_SCK_PORT PORTB
  #define SWSPI_SCK (1 << 8)


#else
  #error "SPI type is not defined. define one of the SPI_HARDWARE, SPI_USART, SPI_SOFTWARE in spi.h"
#endif


// Инициализация SPI-интерфейса
void spi_init();

// Передаёт и принимает 1 байт по SPI, возвращает полученное значение
uint8_t spi_send_recv(uint8_t data);


#endif /* SPI_H_ */